package be.iminds.ilabt.jfed.ui.cli2.instruction;

import com.fasterxml.jackson.annotation.*;
import com.fasterxml.jackson.core.JsonProcessingException;
import com.fasterxml.jackson.databind.ObjectMapper;
import com.fasterxml.jackson.databind.annotation.JsonTypeIdResolver;
import org.slf4j.LoggerFactory;

import javax.annotation.Nonnull;
import javax.annotation.Nullable;
import java.util.Arrays;
import java.util.Collections;
import java.util.List;

/**
 * Instruction: Instructions on what action the CLI has to perform + details needed for it
 */
@JsonTypeInfo(
        use = JsonTypeInfo.Id.CUSTOM,
        include = JsonTypeInfo.As.PROPERTY,
        property = "action",
        defaultImpl = JsonTypeInfo.class, //no default!
        visible = true)
@JsonTypeIdResolver(InstructionTypeIdResolver.class)
//use = JsonTypeInfo.Id.NAME was too limited: it doesn't take enum parsing into account...
//@JsonSubTypes({
//        @JsonSubTypes.Type(value = RunInstruction.class, name = "RUN"),
//        @JsonSubTypes.Type(value = RenewInstruction.class, name = "RENEW"),
//        @JsonSubTypes.Type(value = PoaInstruction.class, name = "POA"),
//        @JsonSubTypes.Type(value = DeleteInstruction.class, name = "DELETE"),
//        @JsonSubTypes.Type(value = CreateSliceInstruction.class, name = "CREATESLICE"),
//        @JsonSubTypes.Type(value = SliceInfoInstruction.class, name = "SLICEINFO"),
//        @JsonSubTypes.Type(value = ManifestInstruction.class, name = "MANIFEST"),
//        @JsonSubTypes.Type(value = CredentialInstruction.class, name = "SLICECREDENTIAL"),
//        @JsonSubTypes.Type(value = UserInfoInstruction.class, name = "USERINFO")
//})
@JsonInclude(JsonInclude.Include.NON_DEFAULT)
@JsonIgnoreProperties(value={"@context"})
public abstract class Instruction {
    private static final org.slf4j.Logger LOG = LoggerFactory.getLogger(Instruction.class);
    protected static final ObjectMapper MAPPER = new ObjectMapper();

    public enum Action {
        //NOTE: keep aliases limited, to keep things as simple as possible
        RUN(true, "run"),

        CREATESLICE(true, "createSlice"),
        POA(true, "performOperationalAction", "poa"),
        RENEW(true, "renew"),
        DELETE(true, "delete"),

        USERINFO(false, "userinfo"),
        SLICEINFO(true, "sliceinfo"),
        SLICECREDENTIAL(true, "sliceCredential"),
        MANIFEST(true, "manifest"),

        DECRYPTPEM(false, "decrypt");

        @Nonnull private final String niceName;
        @Nonnull private final List<String> aliases;
        private final boolean requireSlice;

        Action(boolean requireSlice, @Nonnull String niceName, String... aliases) {
            this.requireSlice = requireSlice;
            this.niceName = niceName;
            this.aliases = Arrays.asList(aliases);
        }

        @JsonCreator
        @Nonnull
        public static Action createFromString(@Nonnull String val) {
            for (Action action : Action.values()) {
                if (action.matches(val))
                    return action;
            }
            throw new RuntimeException("Unknown action: \""+val+"\"");
        }

        @Nonnull
        public static String canonize(@Nonnull String s) {
            return s.trim().replaceAll("[-_ ]?", "").toLowerCase();
        }

        public boolean matches(@Nonnull String cliAction) {
            String canonizedCliAction = canonize(cliAction);
            if (canonize(niceName).equalsIgnoreCase(canonizedCliAction))
                return true;
            for (String alias : aliases) {
                if (canonize(alias).equalsIgnoreCase(canonizedCliAction))
                    return true;
            }
            return false;
        }

        @Nonnull
        public List<String> getAliases() {
            return aliases;
        }

        @Nonnull
        public String getNiceName() {
            return niceName;
        }

        public boolean isSliceArgSupported() {
            return requireSlice;
        }

        @Nullable
        public static Action parse(String cliCommand) {
            for (Action action : Action.values()) {
                if (action.matches(cliCommand))
                    return action;
            }
            return null;
        }
    }

    @Nonnull private final Action action;
    @Nonnull private final User user;

    private final boolean interactive;

    @Nonnull private final List<ActionOutput> actionOutputs;
    @Nonnull private final List<DebugOutput> debugOutputs;
    @Nonnull private final List<CallOutput> callOutputs;

    @JsonCreator
    public Instruction(@JsonProperty("action") @Nonnull Action action,
                       @JsonProperty("user") @Nullable User user,
                       @JsonProperty("interactive") @Nullable Boolean interactive,
                       @JsonProperty("actionOutputs") @Nullable List<ActionOutput> actionOutputs,
                       @JsonProperty("debugOutputs") @Nullable List<DebugOutput> debugOutputs,
                       @JsonProperty("callOutputs") @Nullable List<CallOutput> callOutputs) {
        assert action == getActionOfClass() : action+" != "+getActionOfClass();
        this.action = action;
        this.user = user == null ? new User() : user;
        this.interactive = interactive == null ? true : interactive;
        this.actionOutputs = actionOutputs == null ? Collections.emptyList() : actionOutputs;
        this.debugOutputs = debugOutputs == null ? Collections.emptyList() : debugOutputs;
        this.callOutputs = callOutputs == null ? Collections.emptyList() : callOutputs;
    }

    public Instruction(@Nonnull Action action) {
        this(action, null, null, null, null, null);
    }

    @JsonProperty
    @Nonnull
    public Action getAction() {
        return action;
    }

    @JsonIgnore
    @Nonnull
    protected abstract Action getActionOfClass();


    @JsonProperty
    public boolean isInteractive() {
        return interactive;
    }


    @JsonProperty
    @Nonnull
    public User getUser() {
        return user;
    }

    @JsonProperty
    @Nonnull
    public List<ActionOutput> getActionOutputs() {
        return actionOutputs;
    }

    @JsonProperty
    @Nonnull
    public List<DebugOutput> getDebugOutputs() {
        return debugOutputs;
    }

    @JsonProperty
    @Nonnull
    public List<CallOutput> getCallOutputs() {
        return callOutputs;
    }


//for POA
    @JsonIgnore
    public static final List<String> ACTION_OPTIONS = Arrays.asList("ReloadOS", "Restart", "ConsoleUrl");
    @JsonIgnore
    public static List<String> getActionOptions() {
        return ACTION_OPTIONS;
    }


    public String toString() {
        try {
            return MAPPER.writerWithDefaultPrettyPrinter().writeValueAsString(this);
        } catch (JsonProcessingException e) {
            LOG.error("Exception converting Instruction " + action + " to JSON", e);
            return "Exception converting Instruction " + action + " to JSON: " + e.getMessage();
        }
    }
}
