package be.iminds.ilabt.jfed.rspec.model.javafx_impl;

import be.iminds.ilabt.jfed.lowlevel.api.user_spec.UserSpec;
import be.iminds.ilabt.jfed.rspec.model.*;
import be.iminds.ilabt.jfed.rspec.model.controller.ExperimentBarrierSegment;
import be.iminds.ilabt.jfed.rspec.model.controller.ExperimentCommand;
import be.iminds.ilabt.jfed.rspec.model.impl.*;
import be.iminds.ilabt.jfed.rspec.parser.*;
import be.iminds.ilabt.jfed.rspec.request.ext.delay.SliverTypeShaping;
import be.iminds.ilabt.jfed.util.GeniUrn;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import javax.annotation.Nonnull;
import javax.annotation.Nullable;
import javax.xml.bind.JAXBContext;
import javax.xml.bind.JAXBElement;
import javax.xml.bind.JAXBException;
import javax.xml.bind.Unmarshaller;
import javax.xml.namespace.QName;
import javax.xml.stream.XMLEventReader;
import javax.xml.stream.XMLEventWriter;
import javax.xml.stream.XMLStreamException;
import java.util.*;

import static be.iminds.ilabt.jfed.rspec.parser.RspecXmlConstants.*;

public class FXRspecFactory implements RspecFactory {
    public static final QName Q_SLIVER_TYPE_SHAPING =
            new QName("http://www.protogeni.net/resources/rspec/ext/delay/1", "sliver_type_shaping");
    private final static Logger LOG = LoggerFactory.getLogger(FXRspecFactory.class);
    private static final FXRspecFactory INSTANCE = new FXRspecFactory();

    private FXRspecFactory() {
    }

    public static FXRspecFactory getInstance() {
        return INSTANCE;
    }

    @Nonnull
    @Override
    public ModelRspecType getModelRspecType() {
        return ModelRspecType.FX;
    }

    @Nonnull
    @Override
    public FXRspecNode createNode(@Nonnull ModelRspec modelRspec) {
        assert modelRspec == null || modelRspec instanceof FXModelRspec;
        return new FXRspecNode((FXModelRspec) modelRspec, null);
    }

    @Nonnull
    @Override
    public FXRspecNode createNodeWithClientId(@Nonnull ModelRspec modelRspec, @Nullable String id) {
        assert modelRspec == null || modelRspec instanceof FXModelRspec;
        return new FXRspecNode((FXModelRspec) modelRspec, id);
    }

    @Nonnull
    @Override
    public FXRspecNode copyNode(@Nonnull ModelRspec newModelRspec, @Nonnull RspecNode rspecNode, @Nonnull RspecNode.InterfaceCopyMethod interfaceCopyMethod) {
        assert rspecNode instanceof FXRspecNode;
        assert newModelRspec instanceof FXModelRspec;
        return new FXRspecNode((FXModelRspec) newModelRspec, rspecNode, interfaceCopyMethod);
    }

    @Nonnull
    @Override
    public FXRspecNode copyNodeIncludingInterfaces(@Nonnull ModelRspec newModelRspec, @Nonnull RspecNode rspecNode) {
        assert rspecNode instanceof FXRspecNode;
        assert newModelRspec instanceof FXModelRspec;
        return new FXRspecNode((FXModelRspec) newModelRspec, rspecNode, RspecNode.InterfaceCopyMethod.CREATE_NEW_INTERFACE);
    }

    @Nonnull
    @Override
    public GeantTestbedType copyGeantTestbedType(@Nonnull ModelRspec modelRSpec, @Nonnull GeantTestbedType gtt) {
        return new FXGeantTestbedType(gtt);
    }

    @Nonnull
    @Override
    public RspecNode convertToSpecialNode(@Nonnull RspecNode origNode) throws RspecParseException {
        assert origNode != null;
        assert origNode instanceof FXRspecNode;

        EventListExtraXml delayShapingXml = null;
        if (origNode.getSliverTypes().size() == 1)
            for (ExtraXml extraXml : origNode.getSliverTypes().get(0).getExtraXml()) {
                if (Objects.equals(Q_SLIVER_TYPE_SHAPING, extraXml.getStartElementName())) {
                    assert delayShapingXml == null;
                    assert extraXml instanceof EventListExtraXml;
                    delayShapingXml = (EventListExtraXml) extraXml;
                }
            }

        boolean isDelayBridge = origNode.getSliverTypes().size() == 1 &&
                Objects.equals(origNode.getSliverTypes().get(0).getName(), "delay");

        if (isDelayBridge) {
            //remove ExtraXml
            SliverType origSliverType = origNode.getSliverTypes().get(0);
            List<ExtraXml> sliverTypeExtraXml = new ArrayList<>(origSliverType.getExtraXml());
            sliverTypeExtraXml.remove(delayShapingXml);
            SliverType newSliverType =
                    new SliverTypeBuilder(origSliverType.getName()).setDiskImages(origSliverType.getDiskImages()).setExtraXml(sliverTypeExtraXml).build();
            origNode.getSliverTypes().set(0, newSliverType);

            DelayRspecNode res =
                    new DelayRspecNode((FXModelRspec) origNode.getRspec(), (FXRspecNode) origNode, RspecNode.InterfaceCopyMethod.STEAL_INTERFACE);
            if (delayShapingXml != null) {
                try {
                    JAXBContext jc = JAXBContext.newInstance(SliverTypeShaping.class);
                    Unmarshaller unmarshaller = jc.createUnmarshaller();
                    XMLEventReader eventReader = delayShapingXml.getAsEventReader();
                    JAXBElement<SliverTypeShaping> jb = unmarshaller.unmarshal(eventReader, SliverTypeShaping.class);
                    eventReader.close();

                    SliverTypeShaping sliverTypeShaping = jb.getValue();
                    res.setPropertiesFromSliverTypeShaping(sliverTypeShaping);
                } catch (XMLStreamException | JAXBException e) {
                    throw new RspecParseException(delayShapingXml.getLocation(), "Error parsing SliverTypeShaping: " + e.getMessage(), e);
                }
            }
            return res;
        }

        return origNode;
    }

    @Nonnull
    @Override
    public FXRspecLink createLink(@Nonnull ModelRspec rspec, @Nullable String addDefaultLinkType) {
        assert rspec instanceof FXModelRspec;
        return new FXRspecLink((FXModelRspec) rspec, null, addDefaultLinkType);
    }

    @Nonnull
    @Override
    public FXRspecLink createLinkWithClientId(@Nonnull ModelRspec rspec, @Nonnull String id, @Nullable String addDefaultLinkType) {
        assert rspec instanceof FXModelRspec;
        return new FXRspecLink((FXModelRspec) rspec, id, addDefaultLinkType);
    }

    @Nonnull
    @Override
    public FXRspecLink copyLink(@Nonnull ModelRspec newModelRspec, @Nonnull RspecLink rspecLink, @Nonnull RspecNode.InterfaceCopyMethod interfaceCopyMethod) {
        assert newModelRspec instanceof FXModelRspec;
        return new FXRspecLink((FXModelRspec) newModelRspec, rspecLink, interfaceCopyMethod);
    }

    @Nonnull
    @Override
    public FXRspecInterface createInterface(@Nonnull RspecNode node, @Nullable String id) {
        assert node instanceof FXRspecNode;
        return new FXRspecInterface((FXRspecNode) node, id);
    }

    @Nonnull
    @Override
    public FXRspecInterface createInterface(@Nonnull RspecLink link, @Nullable String id) {
        assert link instanceof FXRspecLink;
        return new FXRspecInterface((FXRspecLink) link, id);
    }

    @Nonnull
    @Override
    public FXRspecInterface createInterface(@Nonnull RspecNode node, @Nonnull RspecLink link, @Nullable String id) {
        assert node instanceof FXRspecNode;
        assert link instanceof FXRspecLink;
        return new FXRspecInterface((FXRspecNode) node, (FXRspecLink) link, id);
    }

    @Nonnull
    @Override
    public FXRspecInterface createInterface(@Nonnull RspecNode node, @Nonnull RspecLink link) {
        assert node instanceof FXRspecNode;
        assert link instanceof FXRspecLink;
        assert node.getRspec() == link.getRspec();
        assert node.getRspec() instanceof FXModelRspec;
        String newClientId = node.getRspec().nextIfaceName(node);
        return new FXRspecInterface((FXRspecNode) node, (FXRspecLink) link, newClientId);
    }

    @Nonnull
    @Override
    public FXLinkSetting createLinkSetting(@Nonnull RspecInterface fromIface, @Nonnull RspecInterface toIface) {
        assert fromIface instanceof FXRspecInterface;
        assert toIface instanceof FXRspecInterface;
        return new FXLinkSetting((FXRspecInterface) fromIface, (FXRspecInterface) toIface);
    }

    @Nonnull
    @Override
    public ExecuteService createExecuteService(@Nullable String shell, @Nullable String command) {
        return new FXExecuteService(shell, command);
    }

    @Nonnull
    @Override
    public InstallService createInstallService(@Nullable String installPath, @Nullable String url) {
        return new FXInstallService(installPath, url);
    }

    @Nonnull
    @Override
    public AnsibleService createAnsibleService() {
        return new BasicAnsibleService();
    }

    @Nonnull
    @Override
    public FXModelRspec createModelRspec(@Nullable String rspecType) {
        return new FXModelRspec(rspecType);
    }

    @Nonnull
    @Override
    public FXModelRspec copyModelRspec(@Nonnull ModelRspec orig, @Nullable String newType) {
        //////////////////////////////////////////////////////////////////
        //WARNING: This is cut-and-paste reuse from BasicRspecFactory!
        //////////////////////////////////////////////////////////////////
        FXModelRspec res = createModelRspec(newType == null ? orig.getType() : newType);

        res.setParseData(orig.getSchemaLocation(), orig.getOrigDefaultNamespace(), orig.getNameSpaceInfo());

        for (RspecNode n : orig.getNodes()) {
            res.addNode(new FXRspecNode(res, n, RspecNode.InterfaceCopyMethod.NO_COPY));
        }

        //copy unbound interfaces
        for (RspecNode n : orig.getNodes()) {
            FXRspecNode copiedNode = res.getNodeByUniqueId(n.getUniqueId());
            assert copiedNode != null;
            if (copiedNode != null) {
                for (RspecInterface iface : n.getInterfaces())
                    if (!iface.isLinkBound()) {
                        //clone interface
                        FXRspecInterface ifaceCopy =
                                new FXRspecInterface(copiedNode, iface.getClientId()); //since we copy to new model, just copy interface name
                        ifaceCopy.overwritePropertiesWith(iface, true);
                    }
            }
        }

        for (RspecLink l : orig.getLinks()) {
            FXRspecLink newLink = new FXRspecLink(res, l, RspecNode.InterfaceCopyMethod.CREATE_NEW_INTERFACE);
            res.addLink(newLink);
            for (RspecInterface i : l.getInterfaces()) {
                FXRspecInterface newIface = newLink.getInterfaceByClientId(i.getClientId());
                FXRspecNode newNode = i.getNode() == null ? null : res.getNodeByUniqueId(i.getNode().getUniqueId());
                assert newIface != null;
                if (newIface != null) {
                    newIface.bindNode(newNode);
                    newIface.overwritePropertiesWith(i, true);
                }
            }
        }

        for (OpenflowDataPath odp : orig.getOpenflowDataPaths()) {
            OpenflowDataPath newOdp =
                    new BasicOpenflowDataPath(odp.getAsExtraXml(), odp.getComponentManagerId(), odp.getComponentId(), odp.getDpid());
            res.addOpenflowDataPath(newOdp);
        }

        for (Lease l : orig.getLeases()) {
            res.addLease(new BasicLease(l));
        }

        for (Channel channel : orig.getChannels()) {
            res.addChannel(new FXRspecChannel(channel));
        }

        for (GeantTestbedType gtt : orig.getGeantTestbedTypes()) {
            res.addGeantTestbedType(new FXGeantTestbedType(gtt));
        }

        for (AddressPool addressPool : orig.getAddressPools()) {
            res.addAddressPool(new FXAddressPool(addressPool));
        }


        if (orig instanceof FXModelRspec) {
            for (ExperimentBarrierSegment bar : ((FXModelRspec) orig).getBarrierSegments()) {
                ExperimentBarrierSegment copyBar = new ExperimentBarrierSegment(bar.getTag(), bar.getOrderNumber());
                //copy commands is done by finishParsing
                res.addBarrierSegment(copyBar);
            }
        }

        List<ExtraXml> extraXmlCopy = new ArrayList<>();
        for (ExtraXml extraXml : orig.getExtraXml()) {
            boolean noStoreNeeded = false;
            try {
                //if this is a FXModelRspec, no extraXml will be processed here, because we will have copied it above already
                noStoreNeeded = this.processExtraRspecXml(res, new EventListExtraXml(extraXml));
            } catch (RspecParseException e) {
                LOG.error("RspecParseException while processing extra node XML: Extra node XML will be ignored", e);
                noStoreNeeded = false;
            }
            if (!noStoreNeeded)
                extraXmlCopy.add(extraXml); //no copy. That should be ok, as only parser will normally change this and nothing after that.
        }
        res.setExtraXml(extraXmlCopy);

        List<UserSpec> userSpecCopy = new ArrayList<>();
        for (UserSpec u : orig.getSshKeys()) {
            userSpecCopy.add(new UserSpec(u));
        }
        res.setSshKeys(userSpecCopy);

        res.setExpires(orig.getExpiresString());

        res.setRoutableAddressesAvailable(orig.getRoutableAddressesAvailable());
        res.setRoutableAddressesConfigured(orig.getRoutableAddressesConfigured());

        //copy command to experiment barrier
        finishParsing(res);

        return res;
        //////////////////////////////////////////////////////////////////
        //WARNING: This is cut-and-paste reuse from BasicRspecFactory!
        //////////////////////////////////////////////////////////////////
    }

    @Override
    public boolean processExtraRspecXml(@Nonnull ModelRspec model, @Nonnull EventListExtraXml extraXml) throws RspecParseException {
        if (Objects.equals(Q_JFEDCOMMAND_EXP_BARRIER_SEG, extraXml.getStartElementName())) {
            ExperimentBarrierSegment barrierSegment =
                    (ExperimentBarrierSegment) StaxRspecParser.parseWithJaxb(extraXml.getStartElement().getLocation(), extraXml.getAsEventReader(), ExperimentBarrierSegment.class);
            ((FXModelRspec) model).getBarrierSegments().add(barrierSegment);
            return true;
        }
        return false;
    }

    @Override
    public boolean processExtraNodeXml(@Nonnull RspecNode node, @Nonnull EventListExtraXml extraXml) throws RspecParseException {
        if (Q_JFEDCOMMAND_EXP_COMMAND.equals(extraXml.getStartElementName())
                || Q_JFEDCOMMAND_EXP_COMMAND_OLDERROR.equals(extraXml.getStartElementName())) {
            ExperimentCommand c = null;
            try {
                c =
                        (ExperimentCommand) StaxRspecParser.parseWithJaxb(extraXml.getStartElement().getLocation(), extraXml.getAsEventReader(), ExperimentCommand.class);
                c.setNodeUniqueID(node.getUniqueId());
                assert c.getNodeUniqueID() != null;
                assert !c.getNodeUniqueID().isEmpty();
                assert c.getTag() != null;
                assert !c.getTag().isEmpty();
                assert c.getStartingTime() != null;
                assert c.getDuration() != null;
            } catch (Exception | AssertionError t) {
                LOG.error("Something went wrong while using JAXB to parse ExperimentCommand: Will ignore ExperimentCommand (and treat it as unknown extra XML).", t);

                //say we don't know this, so it is still stored in ModelRspec
                return false;
            }
            assert c != null;
            assert node instanceof FXRspecNode;
            ((FXRspecNode) node).getCommands().add(c);
            assert node.getUniqueId().endsWith(c.getNodeUniqueID());

            return true;
        }
        return false;
    }

    @Override
    public boolean processExtraLinkXml(@Nonnull RspecLink link, @Nonnull EventListExtraXml extraXml) throws RspecParseException {
        return false;
    }

    @Override
    public void writeExtraRspecXml(@Nonnull ModelRspec aModelRspec, @Nonnull XMLEventWriter eventWriter) {
        FXModelRspec modelRspec = (FXModelRspec) aModelRspec;
        List<ExperimentBarrierSegment> barrierSegments = modelRspec.getBarrierSegments();
        if (barrierSegments != null)
            for (ExperimentBarrierSegment barrierSegment : barrierSegments) {
                StaxRspecWriter.writeWithJaxb(eventWriter, Q_JFEDCOMMAND_EXP_BARRIER_SEG, barrierSegment, ExperimentBarrierSegment.class);
            }
    }

    @Override
    public void writeExtraNodeXml(@Nonnull RspecNode node, @Nonnull XMLEventWriter eventWriter) {
        FXRspecNode fxRspecNode = (FXRspecNode) node;
        for (ExperimentCommand c : fxRspecNode.getCommands()) {
            StaxRspecWriter.writeWithJaxb(eventWriter, Q_JFEDCOMMAND_EXP_COMMAND, c, ExperimentCommand.class);
        }
    }

    @Override
    public void writeExtraLinkXml(@Nonnull RspecLink link, @Nonnull XMLEventWriter eventWriter) {

    }


    @Override
    public boolean equals(Object o) {
        if (!(o instanceof FXRspecFactory)) return false;
        return true;
    }

    @Override
    public int hashCode() {
        return FXRspecFactory.class.hashCode();
    }

    @Override
    public void mergeExtra(@Nonnull ModelRspec aMergedModel, @Nonnull ModelRspec aPartialRspec) {
        FXModelRspec mergedModel = (FXModelRspec) aMergedModel;
        FXModelRspec partialRspec = (FXModelRspec) aPartialRspec;

        partialRspec.getChannels().forEach(mergedModel::addChannel);
        partialRspec.getLeases().forEach(mergedModel::addLease);

        Set<Integer> mergedOrderNumbers = new HashSet<>();
        for (ExperimentBarrierSegment bs : mergedModel.getBarrierSegments())
            mergedOrderNumbers.add(bs.getOrderNumber());
        for (ExperimentBarrierSegment bs : partialRspec.getBarrierSegments())
            if (!mergedOrderNumbers.contains(bs.getOrderNumber())) {
                mergedModel.addBarrierSegment(new ExperimentBarrierSegment(bs.getTag(), bs.getOrderNumber()));
                mergedOrderNumbers.add(bs.getOrderNumber());
            }

        //make sure the commands are added to the barrier
        finishParsing(aMergedModel);
    }

    @Override
    public void finishParsing(@Nonnull ModelRspec modelRspec) {
        FXModelRspec fxModelRspec = (FXModelRspec) modelRspec;
        for (ExperimentBarrierSegment barrierSegment : fxModelRspec.getBarrierSegments())
            for (FXRspecNode n : fxModelRspec.getNodes())
                for (ExperimentCommand command : n.getCommands())
                    if (barrierSegment.getOrderNumber() == command.getBarrierSegmentOrderNumber())
                        barrierSegment.addCommand(command);
    }

    @Nonnull
    @Override
    public OpenflowDataPath createOpenflowDataPath(@Nonnull ModelRspec modelRspec, @Nonnull EventListExtraXml allXml, @Nullable String componentManagerId, @Nullable String componentId, @Nullable String dpidAttValue) {
        //use BasicOpenflowDataPath, as we do nothing special with it for the moment
        assert modelRspec instanceof FXModelRspec;
        return new BasicOpenflowDataPath(allXml, componentManagerId, componentId, dpidAttValue);
    }

    @Nonnull
    @Override
    public Lease createLease(@Nonnull ModelRspec rspec, @Nullable String leaseId, @Nullable String clientId, @Nullable Date validFromDate, @Nullable Date validUntilDate, @Nullable String validFromString, @Nullable String validUntilString) {
        //use BasicLease, as we do nothing special with it for the moment
        assert (rspec instanceof FXModelRspec);
        return new BasicLease(leaseId, clientId, validFromDate, validUntilDate, validFromString, validUntilString);
    }

    @Nonnull
    @Override
    public Lease createLease(@Nonnull ModelRspec rspec, @Nullable String leaseId, @Nullable String clientId, @Nullable Date validFromDate, @Nullable Date validUntilDate) {
        //use BasicLease, as we do nothing special with it for the moment
        assert (rspec instanceof FXModelRspec);
        return new BasicLease(leaseId, clientId, validFromDate, validUntilDate);
    }

    @Nonnull
    @Override
    public Channel createChannel(@Nonnull ModelRspec modelRspec) {
        return new FXRspecChannel();
    }

    @Nonnull
    @Override
    public GeantTestbedType createGeantTestbedType(@Nullable String name) {
        return new FXGeantTestbedType(name);
    }

    @Nonnull
    @Override
    public AddressPool createAddressPool(@Nonnull ModelRspec modelRspec) {
        assert modelRspec instanceof FXModelRspec;
        return new FXAddressPool();
    }

    @Nonnull
    @Override
    public AddressPool copyAddressPool(@Nonnull ModelRspec modelRspec, @Nonnull AddressPool addressPool) {
        assert modelRspec instanceof FXModelRspec;
        return new FXAddressPool(addressPool);
    }

    @Nonnull
    @Override
    public IPv4 createIPv4(@Nonnull ModelRspec modelRspec) {
        assert modelRspec instanceof FXModelRspec;
        return new BasicIPv4();
    }

    @Nonnull
    @Override
    public IPv4 createIPv4(@Nonnull ModelRspec modelRspec, @Nullable String address, @Nullable String netmask) {
        assert modelRspec instanceof FXModelRspec;
        return new BasicIPv4(address, netmask);
    }

    @Nonnull
    @Override
    public ExecuteAnsiblePlaybook createExecuteAnsibleCookbook(@Nonnull ModelRspec rspec, @Nullable String source, @Nullable String outputFilename) {
        return new BasicExecuteAnsiblePlaybook(source, outputFilename);
    }

    @Nonnull
    @Override
    public DistributeSshKeypair createDistributeSshKeypair(@Nonnull ModelRspec rspec, @Nullable String location, @Nullable GeniUrn userUrn) {
        return new BasicDistributeSshKeypair(location, userUrn);
    }

    @Nonnull
    @Override
    public GeantTestbedType.VpnService createGtsVpnService(String endpoint, String username, String password) {
        return new BasicGeantTestbedType.BasicVpnService(endpoint, username, password);
    }

    @Nonnull
    @Override
    public WebVncService createWebVncService(String providerId, String vncUrl, String username, String password) {
        return new WebVncService(providerId, vncUrl, username,password);
    }
}
